@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the Lima2 project
 #
 # Copyright (c) 2020-2024 Beamline Control Unit, ESRF
@@ -20,14 +18,18 @@ Typical usage example:
 
 from __future__ import annotations
 
+import contextlib
 import json
 import logging
 import uuid
+from collections.abc import Iterator
+
+import gevent
 
 import tango
 from beartype import beartype
+from gevent.lock import BoundedSemaphore
 from jsonschema_default import create_from
-from typing_extensions import Optional
 
 from lima2.client import pipelines
 from lima2.client import state_machine as fsm
@@ -47,13 +49,11 @@ class CommError(RuntimeError):
 class Detector:
     """The main Detector interface."""
 
-    DEFAULT_TANGO_TIMEOUT = 5
-
     def __init__(
         self,
         ctrl_dev: tango.DeviceProxy,
         *recv_devs: tango.DeviceProxy,
-        timeout: int = DEFAULT_TANGO_TIMEOUT,
+        timeout: int,
     ):
         """Construct a Detector object.
 
@@ -73,6 +73,7 @@ class Detector:
 
         self.__ctrl = ctrl_dev
         self.__recvs = list(recv_devs)
+        self._state_event_ids: dict[tango.DeviceProxy, int] = {}
 
         self.ping()
 
@@ -82,9 +83,12 @@ class Detector:
                 f"Receiver MPI rank order does not match client order: {rank_order}"
             )
 
-        self.__params_schema: Optional[dict] = None
+        self.__params_schema: dict | None = None
         self.__fsm = fsm.StateMachine(self.__ctrl, self.__recvs)
-        self.__try_sync_hard()
+        self.__fsm_lock = BoundedSemaphore()
+
+        self._register_state_change_handlers()
+        self.sync_hard()
 
         if _logger.getEffectiveLevel() == logging.DEBUG:
             # Log transition of states
@@ -97,6 +101,25 @@ class Detector:
             d.set_green_mode(tango.GreenMode.Gevent)
             d.set_timeout_millis(timeout * 1000)
 
+    def unregister_events(self):
+        """Unregister all state loggers and tango event handlers.
+
+        This method serves only in specific cases where event handlers must for
+        some reason be unregistered. In a normal use case, a Client is
+        instantiated once per Lima2 instance, and no cleanup is necessary.
+        """
+        for device, event_id in self._state_event_ids.items():
+            device.unsubscribe_event(event_id)
+        self._state_event_ids.clear()
+
+    @contextlib.contextmanager
+    def _locked_fsm(self) -> Iterator[fsm.StateMachine]:
+        self.__fsm_lock.acquire()
+        try:
+            yield self.__fsm
+        finally:
+            self.__fsm_lock.release()
+
     @property
     def det_info(self) -> dict:
         """
@@ -317,18 +340,8 @@ class Detector:
         Group:
             State
         """
-        return self.__fsm.state
-
-    def __try_sync_hard(self):
-        """
-        Try to init the state from a lazy constructed device Proxy
-        (to be used when the DS is not running when Detector is constructed)
-        """
-        try:
-            self.ping()
-            self.sync_hard()
-        except tango.DevFailed:
-            return False
+        with self._locked_fsm() as fsm:
+            return fsm.state
 
     def sync_hard(self):
         """
@@ -337,12 +350,13 @@ class Detector:
         Group:
             State
         """
-        try:
-            self.__fsm.sync()
-        except Exception as e:
-            _logger.warning(f"Unable to sync with server: {e}")
+        with self._locked_fsm() as fsm:
+            try:
+                fsm.sync()
+            except Exception as e:
+                _logger.warning(f"Unable to sync with server: {e}")
 
-        return self.__fsm.state
+            return fsm.state
 
     @property
     def dev_states(self):
@@ -370,7 +384,8 @@ class Detector:
         Group:
             State
         """
-        self.__fsm.register_transition_logger(logger)
+        with self._locked_fsm() as fsm:
+            fsm.register_transition_logger(logger)
 
     def unregister_transition_logger(self, logger):
         """
@@ -382,7 +397,8 @@ class Detector:
         Group:
             State
         """
-        self.__fsm.unregister_transition_logger(logger)
+        with self._locked_fsm() as fsm:
+            fsm.unregister_transition_logger(logger)
 
     @beartype
     def prepare_acq(
@@ -404,12 +420,18 @@ class Detector:
             f"prepare_acq({uuid}, {ctrl_params}, {acq_params}, {proc_params})"
         )
 
-        validate(
-            instance=ctrl_params,
-            schema=self.params_schema[self.ctrl.name()]["acq_params"],
-        )
+        def parallel_exec(ctrl_fn, recv_fn):
+            tasks = [gevent.spawn(ctrl_fn, self.__ctrl)]
+            tasks += [gevent.spawn(recv_fn, i, r) for i, r in enumerate(self.__recvs)]
+            gevent.joinall(tasks, raise_error=True)
 
-        for i, recv in enumerate(self.__recvs):
+        def validate_ctrl(ctrl):
+            validate(
+                instance=ctrl_params,
+                schema=self.params_schema[ctrl.name()]["acq_params"],
+            )
+
+        def validate_recv(i, recv):
             validate(
                 instance=acq_params[i],
                 schema=self.params_schema[recv.name()]["acq_params"],
@@ -418,16 +440,22 @@ class Detector:
             pipeline_class = pipelines.get_class(proc_params[i]["class_name"])
             validate(instance=proc_params[i], schema=pipeline_class.params_schema)
 
+        parallel_exec(validate_ctrl, validate_recv)
+
         # Update parameters on the device servers
-        # TODO: Rename acq_params to ctrl_params
-        self.__ctrl.acq_params = json.dumps(ctrl_params)
-        for i, recv in enumerate(self.__recvs):
+        def update_ctrl(ctrl):
+            # TODO: Rename acq_params to ctrl_params
+            ctrl.acq_params = json.dumps(ctrl_params)
+
+        def update_recv(i, recv):
             recv.acq_params = json.dumps(acq_params[i])
             recv.proc_params = json.dumps(proc_params[i])
 
-        res = self.__fsm.prepare(str(uuid))
+        parallel_exec(update_ctrl, update_recv)
 
-        return res.get()  # Makes prepare_acq() synchronous
+        # Raises and reverts to IDLE on exception
+        with self._locked_fsm() as fsm:
+            fsm.prepare(str(uuid))
 
     def start_acq(self):
         """Start acquisition.
@@ -435,7 +463,11 @@ class Detector:
         Group:
             Control
         """
-        self.__fsm.start()
+
+        self._register_on_end_acq_handlers()
+
+        with self._locked_fsm() as fsm:
+            fsm.start()
 
     def trigger(self):
         """Software trigger.
@@ -443,7 +475,11 @@ class Detector:
         Group:
             Control
         """
-        self.__fsm.trigger()
+        with self._locked_fsm() as fsm:
+            if fsm.state == State.FAULT:
+                _logger.warning("Got trigger() while in FAULT state -> ignoring")
+            else:
+                fsm.trigger()
 
     def stop_acq(self):
         """
@@ -452,7 +488,9 @@ class Detector:
         Group:
             Control
         """
-        self.__fsm.stop()
+        with self._locked_fsm() as fsm:
+            fsm.stop()
+            fsm.close()
 
     def reset_acq(self):
         """
@@ -461,7 +499,18 @@ class Detector:
         Group:
             Control
         """
-        self.__fsm.reset_acq()
+        with self._locked_fsm() as fsm:
+            fsm.reset_acq()
+
+    def abort_acq(self):
+        """
+        Abort current acquisition (call Stop on running devices) and go to FAULT state.
+
+        Group:
+            Control
+        """
+        with self._locked_fsm() as fsm:
+            fsm.abort()
 
     @property
     def ctrl_params(self):
@@ -536,3 +585,124 @@ class Detector:
     #         yield self.get_processing(self, uuid)
     #     finally:
     #         self.erase_processing(uuid)
+
+    def _register_state_change_handlers(self):
+        """Register a "change of acq_state" callback on each device.
+
+        These handlers are responsible for handling receivers entering FAULT state
+        (e.g. when the pipeline throws an exception), and for setting the global state
+        to DISCONNECTED if we receive errors on the tango event channel.
+        """
+
+        disconnected_devices = set()
+
+        def on_acq_state_change(evt: tango.EventData):
+            """Handler for state changes of each individual device"""
+            nonlocal disconnected_devices
+
+            # Handle disconnect
+            if evt.err:
+                if evt.device not in disconnected_devices:
+                    _logger.error(
+                        f"Lost connection to device {evt.device.dev_name()}:\n"
+                        + "\n".join([f"- {err.desc}" for err in evt.errors])
+                        + "\n"
+                    )
+                disconnected_devices.add(evt.device)
+                with self._locked_fsm() as fsm:
+                    fsm.disconnect()
+                return
+
+            # Handle FAULT
+            AcqState = type(evt.device.acq_state)
+            acq_state = AcqState(evt.attr_value.value)
+            _logger.debug(f"{evt.device.dev_name()} changed state to {acq_state.name}")
+
+            if acq_state == AcqState.fault:
+                _logger.error(
+                    f"Device {evt.device.dev_name()} in FAULT state.\n"
+                    f"Reason: '{evt.device.last_error}'\n"
+                )
+                if self.state not in [State.FAULT, State.UNKNOWN]:
+                    _logger.error("Aborting acquisition.")
+                    with self._locked_fsm() as fsm:
+                        fsm.abort()
+
+            # Handle reconnect
+            if evt.device in disconnected_devices:
+                _logger.info(f"{evt.device.dev_name()} seems back up")
+                disconnected_devices.remove(evt.device)
+
+            if self.state == State.DISCONNECTED:
+                _logger.info("Trying to sync...")
+                with self._locked_fsm() as fsm:
+                    try:
+                        fsm.sync()
+                        disconnected_devices = set()
+                        print("Lima2 reconnected.")
+                    except Exception as e:
+                        _logger.warning(f"Sync failed: {e}")
+                    finally:
+                        _logger.info(f"Current state: {fsm.state}")
+
+        self._state_event_ids = {
+            dev: dev.subscribe_event(
+                "acq_state",
+                tango.EventType.CHANGE_EVENT,
+                on_acq_state_change,
+            )
+            for dev in self.devs
+        }
+
+    def _register_on_end_acq_handlers(self):
+        """Register an "end of acquisition" callback on each receiver.
+
+        This handler callback is responsible for calling Close on the control device when
+        all receivers are finished transferring frames.
+        """
+        nb_frames_xferreds = []
+        event_ids = {}
+
+        def on_end_acq(evt):
+            nonlocal nb_frames_xferreds
+
+            # Filter spurious error events (if the device is restarted)
+            if evt.err:
+                _logger.error(
+                    f"Error on {evt.device.dev_name()} 'nb_frames_xferred' event:\n"
+                    + "\n".join([f"- {err.desc}" for err in evt.errors])
+                    + "\n"
+                )
+                for r, e in event_ids.items():
+                    r.unsubscribe_event(e)
+                return
+
+            nb_frames_xferreds.append(evt)
+
+            if len(nb_frames_xferreds) == len(self.recvs):
+                for r, e in event_ids.items():
+                    r.unsubscribe_event(e)
+
+                _logger.debug(f"on_end_acq while {self.state}")
+                with self._locked_fsm() as fsm:
+                    if fsm.state == State.RUNNING:
+                        fsm.close()
+                    elif fsm.state == State.FAULT:
+                        # Acquisition aborted -> do nothing
+                        pass
+                    else:
+                        # Unexpected
+                        _logger.info(f"Acquisition ended while in state {fsm.state}")
+
+                nb_frames_xferreds = []
+
+        # On server side, data_ready_event is pushed by each receiver
+        # at the end of the acquisition loop.
+        event_ids = {
+            r: r.subscribe_event(
+                "nb_frames_xferred",
+                tango.EventType.DATA_READY_EVENT,
+                on_end_acq,
+            )
+            for r in self.recvs
+        }
