@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the Lima2 project
 #
 # Copyright (c) 2020-2024 Beamline Control Unit, ESRF
@@ -18,6 +16,7 @@ import traceback
 from collections.abc import Callable
 from concurrent.futures import ThreadPoolExecutor
 from dataclasses import dataclass
+from enum import Enum, auto
 from uuid import UUID
 
 import numpy as np
@@ -26,7 +25,12 @@ from jsonschema_default import create_from
 
 from lima2.client import progress_counter, utils
 from lima2.client.convert import frame_info_to_shape_dtype
-from lima2.client.devencoded import structured_array
+from lima2.client.devencoded import (
+    dense_frame,
+    sparse_frame,
+    smx_sparse_frame,
+    structured_array,
+)
 from lima2.client.progress_counter import ProgressCounter, SingleCounter
 from lima2.client.topology import (
     RoundRobinTopology,
@@ -40,36 +44,57 @@ from lima2.client.utils import classproperty
 logger = logging.getLogger(__name__)
 
 
+class FrameType(Enum):
+    DENSE = auto()
+    SPARSE = auto()
+    SMX_SPARSE = auto()
+
+
+decoder_by_type: dict[FrameType, Callable] = {
+    FrameType.DENSE: dense_frame.decode,
+    FrameType.SPARSE: sparse_frame.decode,
+    FrameType.SMX_SPARSE: smx_sparse_frame.decode_sparse_frame,
+}
+"""Mapping from frame type to associated decode function"""
+
+
 @dataclass
 class FrameSource:
-    """Specifies the name of a getter and a decode function for a given frame source.
+    """Specifies the name of a getter and a frame type for a given frame source.
 
     From these two attributes, it is possible to fetch a frame from a processing tango
     device and then decode it.
     """
 
     getter_name: str
-    decoder: Callable
+    """Name of the getter method to call on the tango device to fetch the data"""
+
+    frame_type: FrameType
+    """Frame type"""
+
+    saving_channel: str | None
+    """Name of the associated saving_params object in the proc_params struct.
+
+    Can be None for sources without persistency.
+    """
+
+    label: str | None
+    """Name of corresponding frame stream in the processing device: progress_counters, etc."""
+
+    saving_counter_name: str | None = None
+    """Name of corresponding saving field in the progress_counters"""
 
 
 class Pipeline:
     """A base class for all processing pipelines."""
 
-    SAVING_PARAMS_NAMES: list[str]
-    """Names of members of type "saving_params" in the corresponding lima2 "proc_params" struct.
-
-    Definition in child classes is enforced by __init_subclass__().
-    """
-
     FRAME_SOURCES: dict[str, FrameSource]
-    """Map of available frame source names to a corresponding getter name and decode function.
+    """Map of available frame source names to a corresponding FrameSource descriptor.
 
     Definition in child classes is enforced by __init_subclass__().
     """
 
-    DEFAULT_TANGO_TIMEOUT = 20 * 60  #
-
-    BYPRODUCT_CACHE_LIMIT = 500
+    BYPRODUCT_CACHE_LIMIT = 30 * 1000
     """Maximum number of cached entries for each processing by-product.
 
     This value corresponds to a number of frames, not a number of array elements.
@@ -125,14 +150,29 @@ class Pipeline:
         schema: dict = json.loads(schema_json)
         return schema
 
+    @staticmethod
+    def distribute_acq(
+        cls,
+        ctl_params: dict,
+        acq_params: list[dict],
+        proc_params: list[dict],
+        topology_kind: TopologyKind,
+    ) -> tuple[dict, list[dict], list[dict]]:
+        """Initialize pipeline-specific parameters for distributed acquisition
+        It is implemented as static method so base classes can access to the
+        derived class implemtation
+        """
+        for i, proc in enumerate(proc_params):
+            # Assign unique filename rank per receiver
+            for source in cls.FRAME_SOURCES.values():
+                if source.saving_channel is not None:
+                    proc[source.saving_channel]["filename_rank"] = i
+
+        return ctl_params, acq_params, proc_params
+
     @classmethod
     def __init_subclass__(cls) -> None:
         """Initialize a pipeline subclass."""
-        if not hasattr(cls, "SAVING_PARAMS_NAMES"):
-            raise ValueError(
-                f"Pipeline subclass {cls} must define a SAVING_PARAMS_NAMES class member"
-            )
-
         if not hasattr(cls, "FRAME_SOURCES"):
             raise ValueError(
                 f"Pipeline subclass {cls} must define a FRAME_SOURCES class member"
@@ -148,7 +188,7 @@ class Pipeline:
         uuid: UUID,
         proc_devs: list[tango.DeviceProxy],
         topology_kind: TopologyKind,
-        timeout: int = DEFAULT_TANGO_TIMEOUT,
+        timeout: int,
     ):
         """Construct a Pipeline object.
 
@@ -165,7 +205,7 @@ class Pipeline:
         self._devs = proc_devs
         self.topology: Topology
 
-        self.byproduct_cache: dict[str, np.ndarray] = {}
+        self.byproduct_cache: dict[str, tuple[int, np.ndarray]] = {}
         """Local cache of processing byproducts.
 
         Used in pop_roi_statistics() and pop_roi_profiles() to return
@@ -173,6 +213,11 @@ class Pipeline:
         speed.
         """
 
+        self.event_info: dict[
+            str, tuple[list[Callable], dict[tango.DeviceProxy, int]]
+        ] = {}
+        """Dictionary holding the Tango event IDs per attribute, per receiver & the callbacks"""
+
         if topology_kind == TopologyKind.SINGLE:
             self.topology = SingleTopology()
 
@@ -216,58 +261,56 @@ class Pipeline:
         ]
 
     def get_frame(self, frame_idx: int, source: str = "frame"):
-        """Get and decode frame by index given source specifier."""
+        """Get and decode frame by index given source specifier.
+
+        Raises:
+          - ValueError: if the source is invalid.
+          - RuntimeError: if the latest frame is requested but no receiver has any frames.
+          - RuntimeError: if fetching the frame raises a DevFailed exception.
+
+        """
         try:
             frame_source = self.FRAME_SOURCES[source]
-        except KeyError as e:
+        except KeyError:
             raise ValueError(
-                f"Invalid source name {e}. "
+                f"Invalid source name {source}. "
                 f"Available ones are: {list(self.FRAME_SOURCES.keys())}"
             )
 
+        try:
+            decode = decoder_by_type[frame_source.frame_type]
+        except KeyError as e:
+            raise NotImplementedError(
+                f"Frame type {frame_source.frame_type} decoder missing"
+            ) from e
+
+        num_rcvs = len(self._devs)
+
+        if num_rcvs == 1 or frame_idx >= 0:
+            # In single-receiver topology or multi-receiver with a valid index
+            # just pass the request to the topology layer
+            effective_frame_idx = frame_idx
+        else:
+            # If asked for the last frame in multi-receiver, try to get its index
+            effective_frame_idx = self.get_last_frame_idx(source=frame_source)
+            if effective_frame_idx < 0:
+                raise RuntimeError(
+                    f"Cannot get the latest <{source}> frame: no frames available from any receiver"
+                )
+
+        logger.debug(f"Getting <{source}> frame: {effective_frame_idx}")
+
         getters = [getattr(dev, frame_source.getter_name) for dev in self._devs]
 
-        if frame_idx == -1:
-            # The most naive way to get the latest frame is to try all receivers.
-            # For monitoring purposes (~1 Hz), this is OK, but would be very wasteful otherwise.
-            num_rcvs = len(self._devs)
+        # We should get a valid frame idx for multi-receiver
+        try:
+            frame = self.topology.lookup(frame_idx=effective_frame_idx, getters=getters)
+        except tango.DevFailed:
+            raise RuntimeError(f"Unable to get frame '{source}' {effective_frame_idx}")
+        return decode(frame)
 
-            def fetch_and_decode_last(rcv_idx):
-                """Return the decoded last frame from receiver rcv_idx.
-
-                Returns None if the receiver hasn't yet processed any frames.
-                """
-                try:
-                    return frame_source.decoder(getters[rcv_idx](-1))
-                except tango.DevFailed:
-                    return None
-
-            with ThreadPoolExecutor(max_workers=num_rcvs) as pool:
-                frames = list(
-                    pool.map(
-                        fetch_and_decode_last,
-                        range(num_rcvs),
-                    )
-                )
-            indices = [frame.idx for frame in frames if frame is not None]
-            if len(indices) == 0:
-                raise ValueError(
-                    "Cannot get the latest frame: no frames available from any receiver"
-                )
-            latest_idx = indices.index(max(indices))
-
-            logger.debug(
-                f"Got latest frame ({max(indices)}) from receiver {latest_idx}"
-            )
-            return frames[latest_idx]
-        else:
-            # Get a frame by absolute index -> use topology-specific lookup()
-            frame = self.topology.lookup(frame_idx=frame_idx, getters=getters)
-
-            return frame_source.decoder(frame)
-
-    def cache_byproduct(self, key: str, data: np.ndarray) -> None:
-        """Store `data` into `self.byproduct_cache[key]`.
+    def cache_byproduct(self, key: str, idx_data: tuple[int, np.ndarray]) -> None:
+        """Store `(last_idx, data)` into `self.byproduct_cache[key]`.
 
         Assumes `data` is a structured array with a "frame_idx" column.
 
@@ -275,6 +318,7 @@ class Pipeline:
         entries corresponding to the excess frames will be discarded
         (starting from the oldest data).
         """
+        last_idx, data = idx_data
         frame_indices = np.unique(data["frame_idx"])
         num_frames = frame_indices.size
         excess_frames = num_frames - self.BYPRODUCT_CACHE_LIMIT
@@ -286,9 +330,11 @@ class Pipeline:
             # Filter out data whose frame_idx is smaller than the last excess frame's
             data = data[data["frame_idx"] >= frame_indices[excess_frames]]
 
-        self.byproduct_cache[key] = data
+        self.byproduct_cache[key] = (last_idx, data)
 
-    def get_cached_byproduct(self, key: str, default: np.ndarray) -> np.ndarray:
+    def get_cached_byproduct(
+        self, key: str, default: tuple[int, np.ndarray]
+    ) -> tuple[int, np.ndarray]:
         return self.byproduct_cache.get(key, default)
 
     @property
@@ -305,7 +351,7 @@ class Pipeline:
             ]
         )
 
-    def pop_roi_statistics(self) -> list[tuple[np.ndarray, np.ndarray]]:
+    def pop_roi_statistics(self) -> list[tuple[np.ndarray, np.ndarray]] | None:
         """Fetch new roi statistics from the server, return them listed by roi index.
 
         Returns:
@@ -327,24 +373,24 @@ class Pipeline:
         ]
 
         # Use cached data from previous call
-        cache = self.get_cached_byproduct(
-            "roi_statistics", default=np.array([], dtype=dtype)
+        last_idx, cache = self.get_cached_byproduct(
+            "roi_statistics", default=(-1, np.array([], dtype=dtype))
         )
 
         # Pop stats for new frames from each receiver and concatenate, mixing in cached data.
         # Data comes from server as 1D array with size (num_frames * num_rois)
-        new_data = np.concatenate(
-            [cache.flatten()]
-            + [
-                structured_array.decode(dev.popRoiStatistics(), dtype)
-                for dev in self._devs
-            ]
-        )
+        def pop_data(dev) -> np.ndarray:
+            return structured_array.decode(dev.popRoiStatistics(), dtype)
+
+        num_rcvs = len(self._devs)
+        with ThreadPoolExecutor(max_workers=num_rcvs) as pool:
+            flat_cache = [cache.flatten()]
+            new_data = np.concatenate(flat_cache + list(pool.map(pop_data, self._devs)))
 
         # Guess number of frames from the data
         num_frames = np.unique(new_data["frame_idx"]).size
         if num_frames == 0:
-            return []
+            return None
 
         # Reshape into (num_frames, num_rois)
         new_data = new_data.reshape((num_frames, -1))
@@ -354,18 +400,22 @@ class Pipeline:
 
         frame_indices = new_data["frame_idx"][:, 0]
         frame_order = np.argsort(frame_indices)
-        first_gap = utils.find_first_gap(frame_indices[frame_order])
+        last_idx, first_gap = utils.find_first_gap(last_idx, frame_indices[frame_order])
 
         # Sort data by frame order
         data = new_data[frame_order]
 
         # Cache any data after the first frame gap for later
-        self.cache_byproduct("roi_statistics", data[first_gap:])
+        self.cache_byproduct("roi_statistics", (last_idx, data[first_gap:]))
+
+        # If did not receive the expected frame return empty dataset
+        if first_gap == 0:
+            return None
 
         # Now our data is contiguous
         data = data[:first_gap]
 
-        logger.debug(f"Returning roi stats for frames: {data['frame_idx'][:,0]}")
+        logger.debug(f"Returning roi stats for frames: {data['frame_idx'][:, 0]}")
 
         return [
             (data[:, i]["frame_idx"], data[:, i][["min", "max", "avg", "std", "sum"]])
@@ -388,7 +438,7 @@ class Pipeline:
 
     def pop_roi_profiles(
         self, roi_lengths: list[int]
-    ) -> list[tuple[np.ndarray, np.ndarray]]:
+    ) -> list[tuple[np.ndarray, np.ndarray]] | None:
         """Fetch new roi profiles from the server, return them listed by roi index.
 
         Args:
@@ -417,37 +467,41 @@ class Pipeline:
         logger.debug(f"Fetching {num_rois} roi profiles with {roi_lengths=}")
 
         # Use cached data from previous call
-        cache = self.get_cached_byproduct(
-            "roi_profiles", default=np.array([], dtype=dtype)
+        last_idx, cache = self.get_cached_byproduct(
+            "roi_profiles", default=(-1, np.array([], dtype=dtype))
         )
 
         # Pop profiles for new frames from each receiver and concatenate, mixing in cached data.
         # Data comes from server as 1D array with size (num_frames * sum(roi_lengths))
-        new_data = np.concatenate(
-            [cache.flatten()]
-            + [
-                structured_array.decode(dev.popRoiProfiles(), dtype)
-                for dev in self._devs
-            ]
-        )
+        def pop_data(dev) -> np.ndarray:
+            return structured_array.decode(dev.popRoiProfiles(), dtype)
+
+        num_rcvs = len(self._devs)
+        with ThreadPoolExecutor(max_workers=num_rcvs) as pool:
+            flat_cache = [cache.flatten()]
+            new_data = np.concatenate(flat_cache + list(pool.map(pop_data, self._devs)))
 
         # Reshape into (num_frames, sum(roi_lengths))
         new_data = new_data.reshape((-1, sum(roi_lengths)))
         num_frames = new_data.shape[0]
         if num_frames == 0:
-            return []
+            return None
 
         logger.debug(f"Received roi profile data for {num_frames} frames")
 
         frame_indices = new_data["frame_idx"][:, 0]
         frame_order = np.argsort(frame_indices)
-        first_gap = utils.find_first_gap(frame_indices[frame_order])
+        last_idx, first_gap = utils.find_first_gap(last_idx, frame_indices[frame_order])
 
         # Sort data by frame order
         data = new_data[frame_order]
 
         # Cache any data after the first frame gap for later
-        self.cache_byproduct("roi_profiles", data[first_gap:])
+        self.cache_byproduct("roi_profiles", (last_idx, data[first_gap:]))
+
+        # If did not receive the expected frame return empty dataset
+        if first_gap == 0:
+            return None
 
         # Now our data is contiguous
         data = data[:first_gap]
@@ -473,6 +527,89 @@ class Pipeline:
             for i in range(num_rois)
         ]
 
+    def pop_raw_frame_idx(self, recv_idx) -> np.ndarray:
+        """Pop recv_idx-2-det_idx frame mapping from one receiver
+
+        The structured array we decode and return contains a list
+        of pairs <recv_idx, det_idx> for all the frames acquired since the last call.
+
+        Returns:
+            A ndarray with <recv_idx, det_idx>
+        """
+        dtype = [
+            ("recv_idx", "i4"),
+            ("det_idx", "i4"),
+        ]
+
+        dev = self._devs[recv_idx]
+        recv_frame_idx = structured_array.decode(dev.popFrameIdx(), dtype)
+
+        logger.debug(
+            f"Popped recv {recv_idx} recv_idx-2-det_idx mapping: {recv_frame_idx}"
+        )
+
+        return recv_frame_idx
+
+    def pop_frame_idx(self) -> tuple[np.ndarray, np.ndarray] | None:
+        """Pop recv_idx-2-det_idx frame mapping
+
+        For each receiver, the structured array we receive contains a list
+        of pairs <recv_frame_idx, det_frame_idx> for all the frames acquired since the last call.
+
+        Returns:
+            A tuple (frame_indices, recv_local_indices), where:
+            - frame_indices is a ndarray of detector frame indices
+            - recv_local_indices is a ndarray of the corresponding <recv_idx, local_idx>
+        """
+        new_dtype = [
+            ("frame_idx", "i4"),
+            ("recv_idx", "i4"),
+            ("local_idx", "i4"),
+        ]
+
+        # Use cached data from previous call
+        last_idx, cache = self.get_cached_byproduct(
+            "frame_idx", default=(-1, np.array([], dtype=new_dtype))
+        )
+
+        # Pop mapping for new frames from each receiver and concatenate, mixing in cached data.
+        # Data is popped from server as 1D array, to be converted to new_dtype
+        new_data_list = []
+        for recv_idx, dev in enumerate(self._devs):
+            recv_data = self.pop_raw_frame_idx(recv_idx)
+            new_data = np.zeros((len(recv_data),), new_dtype)
+            new_data["frame_idx"] = recv_data["det_idx"]
+            new_data["local_idx"] = recv_data["recv_idx"]
+            new_data["recv_idx"] = int(dev.name().split("@")[-1])
+            new_data_list.append(new_data)
+
+        new_data = np.concatenate([cache] + new_data_list)
+
+        logger.debug(
+            f"Received recv_idx-2-det_idx mapping for num_frames={len(new_data_list)}"
+        )
+
+        frame_indices = new_data["frame_idx"]
+        frame_order = np.argsort(frame_indices)
+        last_idx, first_gap = utils.find_first_gap(last_idx, frame_indices[frame_order])
+
+        # Sort data by frame order
+        data = new_data[frame_order]
+
+        # Cache any data after the first frame gap for later
+        self.cache_byproduct("frame_idx", (last_idx, data[first_gap:]))
+
+        # If did not receive the expected frame return empty dataset
+        if first_gap == 0:
+            return None
+
+        # Now our data is contiguous
+        data = data[:first_gap]
+
+        logger.debug(f"Returning recv_idx-2-det_idx for frames: {data['frame_idx']}")
+
+        return (data["frame_idx"], data[["recv_idx", "local_idx"]])
+
     @property
     def progress_counters(self) -> dict[str, ProgressCounter]:
         """Get the list of aggregated progress counters"""
@@ -504,6 +641,27 @@ class Pipeline:
 
         return aggregated_pcs
 
+    def get_last_frame_idx(self, source: FrameSource) -> int:
+        """Get the last available frame index from a given source.
+
+        Raises:
+            ValueError: if progress counter associated with the source is missing
+        """
+
+        if source.label is None:
+            raise ValueError(
+                f"Cannot guess last frame index for {source.getter_name}: "
+                "no associated label"
+            )
+
+        def last_frames(rcv_idx: int) -> int:
+            last_frames_str = self._devs[rcv_idx].last_frames
+            return int(json.loads(last_frames_str)[f"{source.label}_idx"])
+
+        num_rcvs = len(self._devs)
+        with ThreadPoolExecutor(max_workers=num_rcvs) as pool:
+            return max(pool.map(last_frames, range(num_rcvs)))
+
     def ping(self):
         """
         Ping all the devices of the system.
@@ -520,59 +678,60 @@ class Pipeline:
         """A list of `is_finished` for each devices."""
         return [dev.is_finished for dev in self._devs]
 
-    def register_on_finished(self, on_finished: Callable[[], None]) -> None:
+    def register_on_finished(self, cbk: Callable[[], None]) -> None:
         """
         Register a callback function to be notified when all pipelines are finished.
 
         Arg:
-            on_finished: A callback called when all receivers are done
+            cbk: A callback called when all receivers are done
               processing frames.
         """
 
         self.finished_devs: set[tango.DeviceProxy] = set()
         """Set of processing devices that have finished."""
 
-        event_ids: dict[tango.DeviceProxy, int] = {}
+        attr_name = "is_finished"
 
-        def device_finished_callback(evt: tango.DataReadyEventData):
-            nonlocal event_ids
+        # Get the event information
+        cb_list, event_ids = self.event_info.setdefault(attr_name, ([], {}))
+        if cbk not in cb_list:
+            cb_list.append(cbk)
+        else:
+            logger.warning(f"{attr_name}: callback {cbk} already registered")
 
+        # Skip registration if already done
+        if event_ids:
+            return
+
+        all_devs = {dev for dev in self._devs}
+
+        def dev_finished(evt: tango.DataReadyEventData):
             if evt.err:
-                logger.error(
-                    f"Error on {evt.device.dev_name()} 'is_finished' event:\n"
-                    + "\n".join([f"- {err.desc}" for err in evt.errors])
-                    + "\n"
-                )
-                for r, e in event_ids.items():
-                    r.unsubscribe_event(e)
+                self._on_tango_event_error(evt)
                 return
 
-            logger.debug(f"{evt.device.dev_name()} is_finished event received")
+            logger.debug(f"{evt.device.dev_name()} {attr_name} event received")
 
             self.finished_devs.add(evt.device)
-            if self.finished_devs == {dev for dev in self._devs}:
-                # All done. Call on_finished and report any exceptions.
+            if self.finished_devs != all_devs:
+                return
+
+            # All done. Call registered callbacks and report any exceptions.
+            for cbk in cb_list:
                 try:
-                    on_finished()
+                    cbk()
                 except Exception:
                     logger.error(
                         "Exception raised in pipeline on_finished callback:\n"
                         f"{traceback.format_exc()}"
                     )
 
-                for dev, event_id in event_ids.items():
-                    # Unsubscribe from the event to stop trying in case of error
-                    logger.debug(f"Unsubscribing from {event_id} on {dev.dev_name()}")
-                    dev.unsubscribe_event(event_id)
-
-        event_ids = {
-            dev: dev.subscribe_event(
-                "is_finished",
+        for dev in self._devs:
+            event_ids[dev] = dev.subscribe_event(
+                attr_name,
                 tango.EventType.DATA_READY_EVENT,
-                device_finished_callback,
+                dev_finished,
             )
-            for dev in self._devs
-        }
 
     @property
     def last_error(self):
@@ -590,34 +749,76 @@ class Pipeline:
             A dict mapping the pipeline instance name with the event id
         """
 
-        event_ids: dict[tango.DeviceProxy, int] = {}
+        attr_name = "last_error"
 
-        def on_error(evt: tango.DataReadyEventData):
-            nonlocal event_ids
+        cb_list, event_ids = self.event_info.setdefault(attr_name, ([], {}))
+        if cbk not in cb_list:
+            cb_list.append(cbk)
+        else:
+            logger.warning(f"{attr_name}: callback {cbk} already registered")
 
+        # Skip registration if already done
+        if event_ids:
+            return
+
+        def dev_error(evt: tango.DataReadyEventData):
             if evt.err:
-                logger.error(
-                    f"Error on {evt.device.dev_name()} 'last_error' event:\n"
-                    + "\n".join([f"- {err.desc}" for err in evt.errors])
-                    + "\n"
-                )
-            else:
+                self._on_tango_event_error(evt)
+                return
+
+            logger.debug(f"{evt.device.dev_name()} {attr_name} event received")
+
+            for cbk in cb_list:
                 try:
                     cbk(evt)
                 except Exception:
                     logger.error(
-                        "Exception raised in pipeline on_error callback:\n"
+                        "Exception raised in pipeline dev_error callback:\n"
                         f"{traceback.format_exc()}"
                     )
 
-            evt.device.unsubscribe_event(event_ids[evt.device])
+            self._unsubscribe_event(evt.device, event_ids[evt.device])
 
-        event_ids = {
-            proc: proc.subscribe_event(
-                "last_error", tango.EventType.DATA_READY_EVENT, on_error
+        for dev in self._devs:
+            event_ids[dev] = dev.subscribe_event(
+                attr_name, tango.EventType.DATA_READY_EVENT, dev_error
             )
-            for proc in self._devs
-        }
+
+    def _on_tango_event_error(self, evt):
+        """Called when an error is reported in the Tango event layer"""
+
+        attr_name = evt.attr_name
+
+        # skip connection lost-related messages
+        connection_lost = any(
+            ["Event channel is not responding anymore" in err for err in evt.errors]
+        )
+        if not connection_lost:
+            logger.error(
+                f"Error on {evt.device.dev_name()} '{attr_name}' event:\n"
+                + "\n".join([f"- {err.desc}" for err in evt.errors])
+                + "\n"
+            )
+        # something went wrong, disconnect all events
+        self.unregister_events()
+
+    def unregister_events(self):
+        active_attrs = list(self.event_info.keys())
+        for attr_name in active_attrs:
+            self._unsubscribe_attr_events(attr_name)
+
+    def _unsubscribe_attr_events(self, attr_name):
+        """Unsubscribe from the Tango events of a given attribute on all devices"""
+        cb_list, event_ids = self.event_info.pop(attr_name, {})
+        if event_ids:
+            logger.debug(f"Unsubscribing {attr_name} events")
+            for dev, event_id in event_ids.items():
+                self._unsubscribe_event(dev, event_id)
+
+    def _unsubscribe_event(self, dev: tango.DeviceProxy, event_id: int):
+        """Unsubscribe from the Tango event"""
+        logger.debug(f"Unsubscribing from {event_id} on {dev.dev_name()}")
+        dev.unsubscribe_event(event_id)
 
     def __repr__(self) -> str:
         return "\n".join(
