@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the Lima2 project
 #
 # Copyright (c) 2020-2024 Beamline Control Unit, ESRF
@@ -18,22 +16,22 @@ legacy_pipeline = legacy_class(...)
 ```
 """
 
-from typing_extensions import Type
-
 from lima2.client.pipeline import Pipeline
 from lima2.client.pipelines.legacy import Legacy
 from lima2.client.pipelines.smx import Smx
 from lima2.client.pipelines.xpcs import Xpcs
+from lima2.client.pipelines.failing import Failing
 
 # Dictionary of all pipelines supported by the client
 pipelines = {
     "LimaProcessingLegacy": Legacy,
     "LimaProcessingSmx": Smx,
     "LimaProcessingXpcs": Xpcs,
+    "LimaProcessingFailing": Failing,
 }
 
 
-def get_class(tango_class_name: str) -> Type[Pipeline]:
+def get_class(tango_class_name: str) -> type[Pipeline]:
     """Get a pipeline class from its tango class name.
 
     Raise with some additional info if the class doesn't exist.
