@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the Lima2 project
 #
 # Copyright (c) 2020-2024 Beamline Control Unit, ESRF
@@ -7,3 +5,5 @@
 
 from lima2.server.control import Control
 from lima2.server.receiver import Receiver
+
+__all__ = ["Control", "Receiver"]
